/*
 * Copyright (c) 2010, 2013, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */

package jdk.nashorn.internal.runtime;

import java.lang.invoke.MethodHandle;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.concurrent.Callable;
import jdk.nashorn.internal.codegen.ObjectClassGenerator;
import jdk.nashorn.internal.objects.Global;
import jdk.nashorn.internal.parser.JSONParser;
import jdk.nashorn.internal.runtime.arrays.ArrayData;
import jdk.nashorn.internal.runtime.arrays.ArrayIndex;
import jdk.nashorn.internal.runtime.linker.Bootstrap;
import jdk.nashorn.internal.scripts.JO;

/**
 * Utilities used by "JSON" object implementation.
 */
public final class JSONFunctions {
    private JSONFunctions() {}

    private static final Object REVIVER_INVOKER = new Object();

    private static MethodHandle getREVIVER_INVOKER() {
        return Context.getGlobal().getDynamicInvoker(REVIVER_INVOKER,
                new Callable<MethodHandle>() {
                    @Override
                    public MethodHandle call() {
                        return Bootstrap.createDynamicInvoker("dyn:call", Object.class,
                            ScriptFunction.class, ScriptObject.class, String.class, Object.class);
                    }
                });
    }

    /**
     * Returns JSON-compatible quoted version of the given string.
     *
     * @param str String to be quoted
     * @return JSON-compatible quoted string
     */
    public static String quote(final String str) {
        return JSONParser.quote(str);
    }

    /**
     * Parses the given JSON text string and returns object representation.
     *
     * @param text JSON text to be parsed
     * @param reviver  optional value: function that takes two parameters (key, value)
     * @return Object representation of JSON text given
     */
    public static Object parse(final Object text, final Object reviver) {
        final String     str     = JSType.toString(text);
        final Global global = Context.getGlobal();
        final JSONParser parser  = new JSONParser(str);
        final Object value;

        try {
            value = parser.parse();
        } catch (final ParserException e) {
            throw ECMAErrors.syntaxError(e, "invalid.json", e.getMessage());
        }

        final Object unfiltered = convert(global, Global.objectPrototype(), value);
        return applyReviver(global, unfiltered, reviver);
    }

    // -- Internals only below this point

    // parse helpers

    // apply 'reviver' function if available
    private static Object applyReviver(final Global global, final Object unfiltered, final Object reviver) {
        if (reviver instanceof ScriptFunction) {
            final ScriptObject root = global.newObject();
            root.addOwnProperty("", Property.WRITABLE_ENUMERABLE_CONFIGURABLE, unfiltered);
            return walk(root, "", (ScriptFunction)reviver);
        }
        return unfiltered;
    }

    // This is the abstract "Walk" operation from the spec.
    private static Object walk(final ScriptObject holder, final Object name, final ScriptFunction reviver) {
        final Object val = holder.get(name);
        if (val instanceof ScriptObject) {
            final ScriptObject     valueObj = (ScriptObject)val;
            final Iterator<String> iter     = valueObj.propertyIterator();

            while (iter.hasNext()) {
                final String key        = iter.next();
                final Object newElement = walk(valueObj, key, reviver);

                if (newElement == ScriptRuntime.UNDEFINED) {
                    valueObj.delete(key, false);
                } else {
                    setPropertyValue(valueObj, key, newElement);
                }
            }
        }

        try {
             // Object.class, ScriptFunction.class, ScriptObject.class, String.class, Object.class);
             return getREVIVER_INVOKER().invokeExact(reviver, holder, JSType.toString(name), val);
        } catch(Error|RuntimeException t) {
            throw t;
        } catch(final Throwable t) {
            throw new RuntimeException(t);
        }
    }

    // Converts collections to JS objects
    @SuppressWarnings("unchecked")
    private static Object convert(final Global global, final ScriptObject objectProto, final Object value) {
        if (value instanceof Map) {

            final Map<String, Object> map = (Map) value;
            final int length = map.size();
            final List<Property> properties = new ArrayList<>(length);
            final Object[] objectSpill = new Object[length];
            final long[] primitiveSpill = new long[length];
            ArrayData arrayData = ArrayData.EMPTY_ARRAY;
            int slot = 0;

            for (final Map.Entry<String, Object> entry : map.entrySet()) {
                final String name = entry.getKey();
                final Object convertedValue = convert(global, objectProto, entry.getValue());
                final int index = ArrayIndex.getArrayIndex(name);
                if (ArrayIndex.isValidArrayIndex(index)) {
                    // array index key
                    final long oldLength = arrayData.length();
                    final long longIndex = ArrayIndex.toLongIndex(index);
                    if (longIndex > oldLength) {
                        if (arrayData.canDelete(oldLength, longIndex - 1, false)) {
                            arrayData = arrayData.delete(oldLength, longIndex - 1);
                        }
                    }
                    arrayData = arrayData.ensure(longIndex).set(index, convertedValue, false);
                } else {
                    // ordinary property key
                    final Class<?> type;
                    if (ObjectClassGenerator.OBJECT_FIELDS_ONLY) {
                        objectSpill[slot] = convertedValue;
                        type = Object.class;
                    } else {
                        type = getType(convertedValue);
                        if (type == Object.class) {
                            objectSpill[slot] = convertedValue;
                        } else {
                            primitiveSpill[slot] = ObjectClassGenerator.pack((Number) convertedValue);
                        }
                    }
                    final Property property = new SpillProperty(name, 0, slot++);
                    property.setType(type);
                    properties.add(property);
                }
            }

            final ScriptObject result = new JO(PropertyMap.newMap(properties), primitiveSpill, objectSpill);
            result.setInitialProto(objectProto);
            result.setArray(arrayData);
            return result;

        } else if (value instanceof List) {

            final List<Object> list = (List) value;
            if (isNumericArray(list)) {
                final double[] values = new double[list.size()];
                int   index = 0;

                for (final Object obj : list) {
                    values[index++] = JSType.toNumber(obj);
                }
                return global.wrapAsObject(values);
            }

            final Object[] values = new Object[list.size()];
            int   index = 0;

            for (final Object elem : list) {
                values[index++] = convert(global, objectProto, elem);
            }

            return global.wrapAsObject(values);
        }

        return value;
    }


    private static Class<?> getType(final Object value) {
        if (value instanceof Integer) {
            return int.class;
        } else if (value instanceof Long) {
            return long.class;
        } else if (value instanceof Double) {
            return double.class;
        } else {
            return Object.class;
        }
    }

    // add a new property if does not exist already, or else set old property
    private static void setPropertyValue(final ScriptObject sobj, final String name, final Object value) {
        final int index = ArrayIndex.getArrayIndex(name);
        if (ArrayIndex.isValidArrayIndex(index)) {
            // array index key
            sobj.defineOwnProperty(index, value);
        } else if (sobj.getMap().findProperty(name) != null) {
            // pre-existing non-inherited property, call set
            sobj.set(name, value, 0);
        } else {
            // add new property
            sobj.addOwnProperty(name, Property.WRITABLE_ENUMERABLE_CONFIGURABLE, value);
        }
    }

    // does the given list represent a numeric array?
    private static boolean isNumericArray(final List<Object> list) {
        for (final Object obj : list) {
            if (!(obj instanceof Number)) {
                return false;
            }
        }
        return true;
    }
}
